@@ -4,92 +4,115 @@ Events module for creating custom types and generating Messages
 
 from datetime import datetime, timedelta
 import json
-import random
-import time
-import types
-import sys
+from random import choice, randint, random
+from sys import stderr
+from time import sleep
+from types import SimpleNamespace
 
-import faker
+from faker import Faker
 from .handlers import Stream
 
-__all__ = ['EventType', 'EventGenerator']
+__all__ = ['Event', 'EventGenerator']
 
 
-class EventType():
+example_event = {
+    'event_time': '',
+    'type': 'example',
+    'event_id': '',
+    'user_id': '',
+    'first_name': '',
+    'last_name': ''
+}
+
+
+def profile_example(self, profile: dict) -> dict:
+    return {
+        'event_time': self.event_time,
+        'event_id': self.event_id,
+        'user_id': profile.id,
+        'first_name': profile.first_name,
+        'last_name': profile.last_name
+    }
+
+
+class Event():
     """
-    Base Class for new Event Types
+    An Event Types that will produced a defined number of events. The event
+    maybe static or dynamic by suppling a function to modify the event.
 
     Parameters
-    ----------
-        limit: int
-            The number of times to process the event
-
-    Attributes
     ----------
         event: dict
-            The base structure used for the Event
+            Base data structure used for events.  All values must be declared
+            for the automatic value updates to occur.
+        profiler: function
+            A function that takes the profile dict and merges the values. The
+            event can be updated within the function, or return the dict for
+            the update to occur automatically.
+        limit: int
+            The number of times to process the event.
     """
 
-    _next_event = None
-    event = {}
-    event_id = 0
-    event_time = None
-
-    def __init__(self, limit: int = None):
+    def __init__(self,
+                 data: dict,
+                 profiler: callable = None,
+                 limit: int = None):
+        self.data = data
+        self.profiler = profiler
         self.limit = limit
+        self.event_id = 0
+        self.event_time = None
+        self._next_events = None
 
     def __call__(self, profile=None) -> dict:
-        try:
-            self.profiled(profile)
-        except NotImplementedError:
-            pass
-        return self.event
+        if callable(self.profiler):
+            returned = self.profiler(self, profile)
+            if returned:
+                self._update_values(self.data, returned)
+        return self.data
 
     def __repr__(self) -> str:
-        return f'{self.__class__.__name__}(limit={self.limit})'
+        return f'{self.__class__.__name__}({self.data}, {self.profiler}, limit={self.limit})'
+
+    def __lshift__(self, other):
+        other.next = self
+        return other
+
+    def __rshift__(self, other):
+        self.next = other
+        return self
 
     @property
     def next(self):
         """
         View the next event, or use a statement to set the event.
         """
-        return self._next_event
+        return self._next_events
 
     @next.setter
-    def next(self, event) -> None:
-        if isinstance(event, EventType):
-            self._next_event = event
+    def next(self, events) -> None:
+        if events is None:
+            self._next_events = None
+        elif isinstance(events, Event):
+            self._next_events = [events]
+        elif not isinstance(events, list):
+            raise TypeError("An Event or list of EventTypes is required")
+        elif all((isinstance(event, Event) for event in events)):
+            self._next_events = events
         else:
-            raise TypeError("Event must be an EventType")
+            raise TypeError("Events must be only Event Type instances")
 
-    def profiled(self, profile: dict) -> None:
+    @staticmethod
+    def _update_values(dict1: dict, dict2: dict) -> None:
         """
-        If implemented the Event Creator will execute this method, and update
-        the event dict on the instance before it is used.
-
-        The profile can be used for adding details to the event, or updated
-        to carry information onward to other events in the sequence.
+        Merges two dictionaries together, where values in the second dictionary
+        will be applied to the first dictionary.
         """
-        raise NotImplementedError
-
-
-class ExampleEvent(EventType):
-    """
-    Example Event if no event is supplied to the Generator
-    """
-    event = {
-        'type': 'example',
-    }
-
-    def profiled(self, profile: dict) -> dict:
-        updates = {
-            'event_time': self.event_time,
-            'event_id': self.event_id,
-            'user_id': profile.id,
-            'first_name': profile.first_name,
-            'last_name': profile.last_name,
-        }
-        self.event.update(updates)
+        for key, value in dict1.items():
+            if isinstance(value, dict) and key in dict2.keys():
+                Event._update_values(dict1[key], dict2[key])
+            if not isinstance(value, dict) and key in dict2.keys():
+                dict1[key] = dict2[key]
 
 
 class EventGenerator():
@@ -109,32 +132,34 @@ class EventGenerator():
             Customised Faker instance otherise one is created
     """
 
-    _dtstamp = None
-    _first_event = ExampleEvent()
-    _state_table = []
     profiles = []
 
     def __init__(self,
                  num_profiles: int = 10,
                  stream: Stream = None,
                  use_profile_file: bool = False,
-                 fake: faker.Faker = None):
+                 fake: Faker = None):
         self.num_profiles = num_profiles
         self.stream = stream if stream else Stream()
+        self.events = Event(example_event, profile_example, 1)
+        self._dtstamp = None
+        self._state_table = []
+        self._total_count = 0
+        self.profile_filename = 'profiles.json'
 
         self.fake = fake if fake and \
-            isinstance(fake, faker.Faker) else faker.Faker()
+            isinstance(fake, Faker) else Faker()
 
         if use_profile_file:
             try:
-                with open('profiles.json') as profiles_file:
+                with open(self.profile_filename) as profiles_file:
                     profiles_dicts = json.loads(profiles_file.read())
-                    self.profiles = [types.SimpleNamespace(**profiles)
+                    self.profiles = [SimpleNamespace(**profiles)
                                      for profiles in profiles_dicts]
             except FileNotFoundError:
                 self.create_profiles()
 
-                with open('profiles.json', 'w') as profiles_file:
+                with open(self.profile_filename, 'w') as profiles_file:
                     profiles_dicts = [vars(item) for item in self.profiles]
                     profiles_file.write(json.dumps(profiles_dicts))
         else:
@@ -146,29 +171,31 @@ class EventGenerator():
         to process the data if available.
         """
 
-        total_count = 0
-
         if not self._state_table:
             self._reset_state_table()
 
         while self._state_table:
-            sindex = random.randint(0, len(self._state_table)-1)
-            pindex = self._state_table[sindex][0]
-            event = self._state_table[sindex][2]
+            sindex = randint(0, len(self._state_table)-1)
+            pindex = self._state_table[sindex]['pindex']
+            eindex = randint(0, len(self._state_table[sindex]['events'])-1)
+
+            remain = self._state_table[sindex]['events'][eindex]['remain']
+            event = self._state_table[sindex]['events'][eindex]['event']
             selected_profile = self.profiles[pindex]
 
             event.event_time = self._dtstamp.isoformat('T') \
                 if self._dtstamp else datetime.now().isoformat('T')
 
-            total_count += 1
+            self._total_count += 1
             event.event_id += 1
             yield event(selected_profile)
 
-            if isinstance(self._state_table[sindex][1], int):
-                self._state_table[sindex][1] -= 1
-                self._process_state_entry(sindex, event)
+            if remain is not None:
+                self._state_table[sindex]['events'][eindex]['remain'] -= 1
+                self._process_state_entry(sindex, eindex)
 
-        print(f'Event limit reached.  {total_count} in total generated')
+        print(f"Event limit reached.  {self._total_count} in total generated",
+              file=stderr)
 
     def create_profiles(self) -> None:
         """
@@ -177,9 +204,9 @@ class EventGenerator():
         result = []
 
         for _ in range(self.num_profiles):
-            gender = random.choice(['M', 'F'])
+            gender = choice(('male', 'female'))
 
-            if gender == 'F':
+            if gender == 'female':
                 first_name = self.fake.first_name_female()
                 middle_name = self.fake.first_name_female()
                 prefix_name = self.fake.prefix_female()
@@ -195,7 +222,7 @@ class EventGenerator():
                       '|{{postcode}}|{{city}}'
             address1 = self.fake.parse(address).split('|')
             profile = {
-                'id': self.fake.unique.random_number(),
+                'id': str(self.fake.unique.random_number()),
                 'uuid': self.fake.uuid4(),
                 'username': self.fake.user_name(),
                 'gender': gender,
@@ -207,8 +234,8 @@ class EventGenerator():
                 'birthdate': self.fake.date_of_birth(minimum_age=18,
                                                      maximum_age=80)
                             .isoformat(),
-                'blood_group': (random.choice(["A", "B", "AB", "O"]) +
-                                random.choice(["+", "-"])),
+                'blood_group': (choice(["A", "B", "AB", "O"]) +
+                                choice(["+", "-"])),
                 'email': f'{first_name}.{last_name}@{self.fake.domain_name()}',
                 'employer': self.fake.company(),
                 'job': self.fake.job(),
@@ -232,22 +259,24 @@ class EventGenerator():
                 'license_plate': self.fake.license_plate(),
             }
 
-            result.append(types.SimpleNamespace(**profile))
+            result.append(SimpleNamespace(**profile))
         self.profiles = result
 
     @property
-    def first_event(self) -> EventType:
+    def events(self) -> list:
         """
         View the first event, or use a statement to set the event.
         """
-        return self._first_event
+        return self._events
 
-    @first_event.setter
-    def first_event(self, event: EventType) -> None:
-        if isinstance(event, EventType):
-            self._first_event = event
+    @events.setter
+    def events(self, events: list) -> None:
+        if isinstance(events, Event):
+            self._events = [events]
+        elif all((isinstance(event, Event) for event in events)):
+            self._events = events
         else:
-            raise TypeError("Events must be an EventType instance")
+            raise TypeError("Events must be only Event Type instances")
 
         self._reset_state_table()
 
@@ -262,9 +291,10 @@ class EventGenerator():
         try:
             for event in self.create_events():
                 self.stream.send(json.dumps(event, indent=indent))
-                time.sleep(random.random() * 60/epm)
+                sleep(random() * 60/epm)
         except KeyboardInterrupt:
-            print('\nStopping Event Stream', file=sys.stderr)
+            print(f"\nStopping Event Stream.  {self._total_count} in total generated.",
+                  file=stderr)
 
     def batch(self,
               start: datetime,
@@ -281,20 +311,45 @@ class EventGenerator():
         try:
             for event in self.create_events():
                 if self._dtstamp >= finish:
-                    print('Finish time reached', file=sys.stderr)
+                    print(f"Finish time reached.  {self._total_count} in total generated.",
+                          file=stderr)
                     break
                 self.stream.send(json.dumps(event, indent=indent))
-                self._dtstamp += timedelta(seconds=random.random() * 60/epm)
+                self._dtstamp += timedelta(seconds=random() * 60/epm)
         except KeyboardInterrupt:
-            print('\nStopping Event Batch', file=sys.stderr)
+            print(f"\nStopping Event Batch.  {self._total_count} in total generated.",
+                  file=stderr)
 
     def _reset_state_table(self) -> None:
-        self._state_table = [[index, self.first_event.limit, self.first_event]
-                             for index, _ in enumerate(self.profiles)]
+        self._state_table = [
+            {
+                'pindex': index,
+                'events': [
+                    {
+                        'remain': event.limit,
+                        'event': event
+                    } for event in self.events
+                ]
+            }
+            for index, _ in enumerate(self.profiles)
+        ]
+        self._total_count = 0
 
-    def _process_state_entry(self, index: int, event: EventType) -> None:
-        if self._state_table[index][1] == 0 and event.next is None:
-            del self._state_table[index]
-        elif self._state_table[index][1] == 0:
-            self._state_table[index][1] = event.next.limit
-            self._state_table[index][2] = event.next
+    def _process_state_entry(self, sindex: int, eindex: int) -> None:
+        remain = self._state_table[sindex]['events'][eindex]['remain']
+        event = self._state_table[sindex]['events'][eindex]['event']
+
+        if remain <= 0 and event.next:
+            del self._state_table[sindex]['events'][eindex]
+            event_record = [
+                {
+                    'remain': next_event.limit,
+                    'event': next_event
+                } for next_event in event.next
+            ]
+            self._state_table[sindex]['events'].extend(event_record)
+        elif remain <= 0:
+            del self._state_table[sindex]['events'][eindex]
+
+        if not self._state_table[sindex]['events']:
+            del self._state_table[sindex]
